import os
import numpy as np
import scipy as sp
import dimod
import minorminer
import gurobipy
from tqdm import tqdm
from concurrent.futures import ThreadPoolExecutor
from dwave.samplers import SimulatedAnnealingSampler
from dwave.system import LeapHybridSampler
from dwave.system import DWaveSampler, FixedEmbeddingComposite


class QUnfoldQUBO:
    def __init__(self, response, measured, lam=0.0):
        self.R = response
        self.d = measured
        self.lam = lam

        dim = len(measured)
        main_d = np.array([0, -1] + [-2] * (dim - 4) + [-1, 0])
        sup_sub_d = np.array([0] + [1] * (dim - 3) + [0])
        self.D = np.diag(main_d) + np.diag(sup_sub_d, k=1) + np.diag(sup_sub_d, k=-1)

    @property
    def num_bins(self):
        return len(self.d)

    @property
    def num_logical_qubits(self):
        return self.dwave_bqm.num_variables

    @property
    def num_physical_qubits(self):
        return sum(len(chain) for chain in self.graph_embedding.values())

    @property
    def num_bits(self):
        eff = np.sum(self.R, axis=0)
        eff[np.isclose(eff, 0)] = 1
        exp = np.ceil(self.d / eff)
        return [int(np.ceil(np.log2(x * 1.2))) if x else 1 for x in exp]

    @property
    def precision_vectors(self):
        num_bits = self.num_bits
        pvecs = [2 ** np.arange(num_bits[i]) for i in range(self.num_bins)]
        return pvecs

    @property
    def linear_coeffs(self):
        return -2 * (self.R.T @ self.d)

    @property
    def quadratic_coeffs(self):
        return (self.R.T @ self.R) + self.lam * (self.D.T @ self.D)

    def _get_qubo_matrix(self):
        dim = self.num_bins
        a = self.linear_coeffs
        B = self.quadratic_coeffs
        pvecs = self.precision_vectors
        linear_blocks = [a[i] * np.diag(pvecs[i]) for i in range(dim)]
        quadratic_blocks = [
            [B[i, j] * np.outer(pvecs[i], pvecs[j]) for j in range(dim)]
            for i in range(dim)
        ]
        return sp.linalg.block_diag(*linear_blocks) + np.block(quadratic_blocks)

    def _get_dwave_bqm(self):
        Q = self.qubo_matrix
        size = len(Q)
        linear = {i: Q[i, i] for i in range(size)}
        quadratic = {
            (i, j): 2 * Q[i, j] for i in range(size) for j in range(i + 1, size)
        }
        return dimod.BinaryQuadraticModel(linear, quadratic, vartype=dimod.BINARY)

    def _post_process_sampleset(self, sampleset):
        pvecs = self.precision_vectors
        indices = np.cumsum([0] + [len(pvec) for pvec in pvecs[:-1]])
        solutions = [
            np.add.reduceat(rec.sample * np.concatenate(pvecs), indices=indices)
            for rec in sampleset.record
        ]
        energies = [rec.energy for rec in sampleset.record]
        min_energy = min(energies)
        beta_boltzmann = 100
        weights = np.exp(-beta_boltzmann * np.abs(np.array(energies) - min_energy))
        return np.average(solutions, weights=weights, axis=0)

    def _get_graph_embedding(self, **kwargs):
        source_edgelist = list(self.dwave_bqm.quadratic) + list(
            (v, v) for v in self.dwave_bqm.linear
        )
        target_edgelist = self._sampler.edgelist
        return minorminer.find_embedding(S=source_edgelist, T=target_edgelist, **kwargs)

    def _run_montecarlo_toys(self, num_toys, num_cores, **kwargs):
        def run_toy(toy):
            toy.initialize_qubo_model()
            if isinstance(self._sampler, DWaveSampler):
                embedding = toy._get_graph_embedding()
                sampler = FixedEmbeddingComposite(self._sampler, embedding=embedding)
            else:
                sampler = self._sampler
            sampleset = sampler.sample(toy.dwave_bqm, **kwargs)
            solution = toy._post_process_sampleset(sampleset)
            return solution

        smeared_d = np.random.poisson(self.d, size=(num_toys, self.num_bins))
        toys_gen = (QUnfoldQUBO(self.R, d, self.lam) for d in smeared_d)
        if num_cores is None:
            num_cores = os.cpu_count() - 2
        desc = "Running MC toys"
        with ThreadPoolExecutor(num_cores) as executor:
            res = list(tqdm(executor.map(run_toy, toys_gen), total=num_toys, desc=desc))
        cov = np.cov(res, rowvar=False)
        err = np.sqrt(np.diag(cov))
        return err, cov

    def initialize_qubo_model(self):
        self.qubo_matrix = self._get_qubo_matrix()
        self.dwave_bqm = self._get_dwave_bqm()

    def solve_simulated_annealing(
        self, num_reads, num_toys=None, num_cores=None, seed=None
    ):
        self._sampler = SimulatedAnnealingSampler()
        sampleset = self._sampler.sample(self.dwave_bqm, num_reads=num_reads, seed=seed)
        sol = self._post_process_sampleset(sampleset)
        if num_toys is None:
            err = np.sqrt(sol)
            cov = np.diag(sol)
        else:
            err, cov = self._run_montecarlo_toys(
                num_toys, num_cores, num_reads=num_reads, seed=seed
            )
        return sol, err, cov

    def solve_hybrid_sampler(self, num_toys=None, num_cores=None):
        self._sampler = LeapHybridSampler()
        sampleset = self._sampler.sample(self.dwave_bqm)
        sol = self._post_process_sampleset(sampleset)
        if num_toys is None:
            err = np.sqrt(sol)
            cov = np.diag(sol)
        else:
            err, cov = self._run_montecarlo_toys(num_toys, num_cores)
        return sol, err, cov

    def set_quantum_device(self, device_name=None, dwave_token=None):
        self._sampler = DWaveSampler(solver=device_name, token=dwave_token)

    def set_graph_embedding(self, **kwargs):
        self.graph_embedding = self._get_graph_embedding(**kwargs)

    def solve_quantum_annealing(self, num_reads, num_toys=None, num_cores=None):
        sampler = FixedEmbeddingComposite(self._sampler, embedding=self.graph_embedding)
        sampleset = sampler.sample(self.dwave_bqm, num_reads=num_reads)
        sol = self._post_process_sampleset(sampleset)
        if num_toys is None:
            err = np.sqrt(sol)
            cov = np.diag(sol)
        else:
            err, cov = self._run_montecarlo_toys(
                num_toys, num_cores, num_reads=num_reads
            )
        return sol, err, cov

    def compute_energy(self, x):
        xbin = []
        num_bits = self.num_bits
        for i in range(self.num_bins):
            bitstr = bin(int(x[i]))[2:].zfill(num_bits[i])
            xbin.extend(map(int, bitstr[::-1]))
        energy = self.dwave_bqm.energy(sample=xbin)
        return energy

    def solve_gurobi_integer(self):
        model = gurobipy.Model()
        model.setParam("OutputFlag", 0)
        x = [
            model.addVar(vtype=gurobipy.GRB.INTEGER, lb=0, ub=2**b - 1)
            for b in self.num_bits
        ]
        a = self.linear_coeffs
        B = self.quadratic_coeffs
        model.setObjective(a @ x + x @ B @ x, sense=gurobipy.GRB.MINIMIZE)
        model.optimize()
        sol = np.array([var.x for var in x])
        err = np.sqrt(sol)
        cov = np.diag(sol)
        return sol, err, cov

    def solve_gurobi_binary(self):
        model = gurobipy.Model()
        model.setParam("OutputFlag", 0)
        num_bits = self.num_bits
        x = [
            model.addVar(vtype=gurobipy.GRB.BINARY)
            for i in range(self.num_bins)
            for _ in range(num_bits[i])
        ]
        Q = self.qubo_matrix
        model.setObjective(x @ Q @ x, sense=gurobipy.GRB.MINIMIZE)
        model.optimize()
        bitstr = np.array([var.x for var in x], dtype=int)
        arrays = np.split(bitstr, np.cumsum(num_bits[:-1]))
        sol = np.array(
            [int("".join(arr.astype(str))[::-1], base=2) for arr in arrays], dtype=float
        )
        err = np.sqrt(sol)
        cov = np.diag(sol)
        return sol, err, cov



import numpy as np


def normalize_response(response, truth_mc):
    """
    Normalize the response matrix using the Monte Carlo generated truth histogram.

    Args:
        response (numpy.ndarray): response matrix to normalize.
        truth_mc (numpy.ndarray): Monte Carlo truth histogram.

    Returns:
        numpy.ndarray: normalized response matrix.
    """
    return response / (truth_mc + 1e-12)


def compute_chi2(observed, expected, covariance):
    """
    Compute the reduced chi-square between the observed and the expected histogram.

    Args:
        observed (numpy.ndarray): observed histogram.
        expected (numpy.ndarray): expected histogram.
        covariance (numpy.ndarray): covariance matrix.

    Returns:
        float: reduced chi-square.
    """
    residuals = observed - expected
    chi2 = residuals.T @ np.linalg.pinv(covariance) @ residuals
    chi2_red = chi2 / len(expected)
    return chi2_red



import numpy as np
import pylab as plt


class QUnfoldPlotter:
    """
    Class used to plot QUnfold data and results.
    """

    def __init__(self, response, measured, truth, unfolded, error, binning, chi2):
        """
        Constructs a QUnfoldPlotter class for visualizing unfolding results.

        Args:
            response (numpy.ndarray): response matrix used in the unfolding process.
            measured (numpy.ndarray): measured data histogram.
            truth (numpy.ndarray): truth histogram.
            unfolded (numpy.ndarray): unfolded histogram.
            error (numpy.ndarray): errors on the unfolded histogram.
            binning (numpy.ndarray): binning of the histograms.
            chi2 (float): chi2 to show on the plot.
        """
        self.response = response[1:-1, 1:-1]
        self.measured = measured[1:-1]
        self.truth = truth[1:-1]
        self.unfolded = unfolded[1:-1]
        self.error = error[1:-1]
        self.binning = binning[1:-1]
        self.chi2 = chi2

    def _plotResponseSetup(self):
        """
        Set up the response matrix plot for drawing or saving: the response matrix is set up as a heatmap,
        with the column representing the measured variable and the row representing the truth variable.
        """
        # Set up plot
        plt.imshow(
            np.transpose(self.response),
            cmap="viridis",
            extent=[
                self.binning[0],
                self.binning[-1],
                self.binning[0],
                self.binning[-1],
            ],
            origin="lower",
        )
        plt.colorbar()
        plt.xlabel("Truth")
        plt.ylabel("Measured")

    def plotResponse(self):
        """
        Plot the response matrix with matplotlib style.
        """
        self._plotResponseSetup()
        plt.show()
        plt.close()

    def saveResponse(self, path):
        """
        Save the plotted response matrix with matplotlib style to a file.

        Args:
            path (str): file path to save the plot.
        """
        self._plotResponseSetup()
        plt.savefig(path)
        plt.close()

    def _plotSetup(self, method):
        """
        Create an histogram comparison among measured, truth and unfolded distributions,
        with the chi2 among unfolded and truth distribution.

        Args:
            method (str): unfolding method.
        """
        # Divide into subplots
        fig = plt.figure()
        gs = fig.add_gridspec(2, 1, height_ratios=[3, 1], hspace=0)
        ax1 = fig.add_subplot(gs[0])
        ax2 = fig.add_subplot(gs[1], sharex=ax1)

        # Plot truth
        truth_steps = np.append(self.truth, [self.truth[-1]])
        ax1.step(
            self.binning, truth_steps, label="Truth", where="post", color="tab:blue"
        )
        ax1.fill_between(
            self.binning, truth_steps, step="post", alpha=0.3, color="tab:blue"
        )

        # Plot measured
        meas_steps = np.append(self.measured, [self.measured[-1]])
        ax1.step(
            self.binning, meas_steps, label="Measured", where="post", color="tab:orange"
        )

        # Plot unfolded histogram with chi2 test
        binwidths = np.diff(self.binning)
        bin_midpoints = self.binning[:-1] + binwidths / 2
        chi2 = round(self.chi2, 2)
        label = rf"Unfolded {method} ($\chi^2 = {chi2}$)"
        ax1.errorbar(
            x=bin_midpoints,
            y=self.unfolded,
            yerr=self.error,
            label=label,
            marker="o",
            ms=3.5,
            c="green",
            linestyle="None",
        )

        # Plot ratio QUnfold to truth
        ax2.axhline(y=1, color="tab:blue")
        ax2.errorbar(
            x=bin_midpoints,
            y=self.unfolded / self.truth,
            yerr=self.error / self.truth,
            ms=3.5,
            fmt="o",
            color="g",
        )

        # Plot style settings
        ax1.tick_params(axis="x", which="both", bottom=True, top=False, direction="in")
        ax2.tick_params(axis="x", which="both", bottom=True, top=True, direction="in")
        ax1.set_xlim(self.binning[0], self.binning[-1])
        ax1.set_ylim(0, ax1.get_ylim()[1])
        ax2.set_yticks([0.25, 0.5, 0.75, 1.0, 1.25, 1.5, 1.75])
        ax2.set_yticklabels(["", "0.5", "", "1.0", "", "1.5", ""])
        ax1.tick_params(
            axis="x", which="both", bottom=False, top=False, labelbottom=False
        )

        # Plot settings
        ax2.set_ylabel("Ratio to\ntruth")
        ax2.set_xlabel("Bins")
        ax1.set_ylabel("Entries")
        ax1.legend(loc="upper right")
        plt.tight_layout()

    def plot(self, method=""):
        """
        Plot the measured distribution histogram.

        Args:
            method (str): unfolding method.
        """
        self._plotSetup(method)
        plt.show()
        plt.close()

    def savePlot(self, path, method=""):
        """
        Save the plot of the measured distribution histogram into path.

        Args:
            path (str): file path to save the plot.
            method (str): unfolding method.
        """
        self._plotSetup(method)
        plt.savefig(path)
        plt.close()



import numpy as np


def TH1_to_numpy(histo, error=False, overflow=False, dtype=np.float64):
    """
    Convert a ROOT.TH1 object into a 1D numpy array.

    Args:
        histo (ROOT.TH1): the input TH1 histogram to convert.
        error (bool): if True, it gets the bin error instead of the bin content.
        overflow (bool): if True, it considers also the overflow bins.
        dtype (type): the data type of the output array.

    Returns:
        array (numpy.ndarray): the output 1D numpy array.
    """
    get_entry = histo.GetBinError if error else histo.GetBinContent
    if overflow:
        irange = range(0, histo.GetNbinsX() + 2)
    else:
        irange = range(1, histo.GetNbinsX() + 1)
    array = np.array([get_entry(i) for i in irange])
    return array.astype(dtype=dtype)


def TH2_to_numpy(histo, overflow=False, dtype=np.float64):
    """
    Convert a ROOT.TH2 object into a 2D numpy array.

    Args:
        histo (ROOT.TH2): the input TH2 histogram to convert.
        overflow (bool): if True, it considers also the overflow bins.
        dtype (type): the data type of the output array.

    Returns:
        array (numpy.ndarray): the output 2D numpy array.

    """
    if overflow:
        irange = range(0, histo.GetNbinsX() + 2)
        jrange = range(0, histo.GetNbinsY() + 2)
    else:
        irange = range(1, histo.GetNbinsX() + 1)
        jrange = range(1, histo.GetNbinsY() + 1)
    array = np.array([[histo.GetBinContent(i, j) for j in jrange] for i in irange])
    return array.astype(dtype=dtype)


def TVector_to_numpy(vector, dtype=np.float64):
    """
    Convert a ROOT.TVector object into a 1D numpy array.

    Args:
        vector (ROOT.TVector): the input TVector to convert.
        dtype (type): the data type of the output array.

    Returns:
        array (numpy.ndarray): the output 1D numpy array.
    """
    num_elements = vector.GetNoElements()
    array = np.array([vector[i] for i in range(num_elements)])
    return array.astype(dtype=dtype)


def TMatrix_to_numpy(matrix, dtype=np.float64):
    """
    Convert a ROOT.TMatrix object into a 2D numpy array.

    Parameters:
        matrix (ROOT.TMatrix): the input TMatrix to convert.
        dtype (type): the data type of the output array.

    Returns:
        array (numpy.ndarray): the output 2D numpy array.
    """
    num_rows = matrix.GetNrows()
    num_cols = matrix.GetNcols()
    array = np.array([[matrix[i][j] for j in range(num_cols)] for i in range(num_rows)])
    return array.astype(dtype=dtype)
